#
# This file is part of Zaplet
# Copyright 1999 Adam Feuer <adamf@pobox.com>
#
# Zaplet is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Zaplet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Zaplet; see the file COPYING-Zaplet. If not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#

# zapreader
# zaplet converter classes
# for use with adzapper zaplets

RCS_ID =  '$Id: zapreader.py,v 1.2 2000/08/11 07:05:18 adamf Exp $'

import os
import os.path
import re
import string
import os
import sys
import urlparse
import socket

import debug

numeric_host = re.compile ('[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+')
numeric_network = re.compile ('[0-9]+\.([0-9]+\.)?([0-9]+\.)?([[0-9]+\*])?')


def quoted_string(arg):
    """returns true if string is enclosed in quotes"""
    l = len(arg)
    if l > 1:
        if arg[0] == "\"" and arg[l-1] == "\"":
            return 1
        elif arg[0] == "\'" and arg[l-1] == "\'":
            return 1
        return 0


def regex_string(arg):
    """returns true of string is enclosed in '/'"""
    l = len(arg)
    if l > 1:
        if arg[0] == '/' and arg[l-1] == '/':
            return 1
    return 0


def trim(arg):
    """cut off the first and last characters of the string"""
    l = len(arg)
    if l > 2:
        return arg[1:l-1]
    else:
        return ""



class zaplet:
    """data structure object to hold zaplet"""
    def __init__(self):
        # zaplet version defaults to 0.1
        self.majorversion = 0
        self.minorversion = 1
        self.host = None
        self.allow = ''
        self.block = ''
        self.lines = []

        self.version = ''
        self.block_list = []
        self.allow_list = []
        self.comment_list = []



class zaplet_converter_engine:
    """Zaplet engine class: this does all the work--
    reads in zaplets, and determines whether URLs should be blocked.

    can be called with string that has a single zaplet directory,
    or with a list that contains multiple zaplet directory paths, each
    as a single string

    """

    def __init__(self, arg):
        """initialize the zaplet_engine object
        build the {site : block_regex} dictionary"""

        debug.debug(1,"Initializing zaplet engine...")

        self.zDirs = []

        # set the list of zaplet directories
        if type(arg) == type(""):
            self.zDirs.append(arg)
        elif type(arg) == type([]):
            self.zDirs = arg

        # dictionary of (site, zaplet object instance) pairs
        self.zaplet_list = { }

        self.readZaplets(self.zDirs)




    def readZaplets(self, zDirs):
        """read zaplets in from each of the directories in the list
        of zaplet directories

        read them in the order specified.

        for zaplets that have the same host statement,
        the most recent zaplet read overwrites ones that have
        already been read in
        
        """
        for dir in zDirs:
            self.readZapletDir(dir)




    def readZapletDir(self,zDir):
        """read Zaplets from a directory
        """
        debug.debug(1,"zaplet_engine- reading zaplets from '" + zDir + "'...")

        cpairs = []  # list of (site, conf file) pairs

        # make the list of zaplets

        if len(zDir) < 1:
            debug.debug(3,"warning: empty zaplet directory string")
            return

        # zDir must end with '/' 
        if zDir[-1] != '/':
            zDir = zDir + '/'

        if (zDir[0] == '/'):
            # it is an absolute path
            pass
        else:
            # not absolute path to zaplet directory-- add cwd to it
            self.zDir = os.path.join(os.getcwd(),zDir)

        zdirlist = []
        try:
            zdirlist = os.listdir(zDir)
        except:
            debug.debug(1,"zaplet_engine- can't read zaplet directory '" + zDir + "'")
            return
        debug.debug(3, "zaplet_engine- zdirlist:")
        debug.debug(3, zdirlist)

        zapfile =  re.compile('^.*\.zap$')
        dotfile =  re.compile('^\.')
        conffile = re.compile('.*\.conf$')
        zlist = []
        for file in zdirlist:
            if os.path.isfile(os.path.join(zDir,file)):
                if zapfile.search(file) and not dotfile.search(file):
                    zlist.append(file)
        debug.debug(3,"zaplet_engine- processed source directory: ")
        debug.debug(3,zlist)

        # now go thru list of zaplets, read them in and compile them.
        for zfile in zlist:
            debug.debug(1,"zaplet_engine- site: " + zfile + "        zaplet file: " + zfile)
            zl = self.get_zaplet(zDir,zfile)

            # if we got a list of zaplets back...
            if len(zl) > 0:
                # add each of the zaplets to the dictionary
                for zlitem in zl:
                    self.zaplet_list[zlitem.host] = zlitem

        debug.debug(1,self.zaplet_list.keys())



    def get_zaplet(self,zDir,zfile):
        """make the list of lists of zaplet lines
        given the name of a zapletfile, read it in, split any zaplets into separate lists of lines,
        and call make_zaplet() on each list of lines.
        """

        zapletfile = os.path.join(zDir,zfile)
        
        zaplist = []

        try:
            debug.debug(1," opening zaplet : " + zapletfile)
            cf = open(zapletfile, "r")
        except:
            sys.stderr.write("Cannot open zaplet " + zapletfile + "\n")
            return None

        lines = cf.readlines()
        cf.close()

        in_zaplet = 0
        for line in lines:
            if in_zaplet:
                zlines.append(line)
                if re.search('^\s*</zaplet>',line):
                    in_zaplet = 0
                    zaplist.append(zlines)
                    debug.debug(1,zlines)
            else: # not in_zaplet
                if re.search('^\s*<zaplet>',line):
                    in_zaplet = 1
                    zlines = []
                    zlines.append(line)

        zaplet_list = []
        for zlitem in zaplist:
            tmp = self.make_zaplet(zlitem,zfile)
            if tmp != None:
                zaplet_list.append(tmp)
            else:
                debug.debug(1,"%s: zaplet not used." % zapletfile)

        return zaplet_list



    def make_zaplet(self,zapletlines,zfile):
        """make a new zaplet
        given the list of lines that make a zaplet, parse the lines and return a zaplet object
        complete with compiled regexes

        returns None if zaplet is invalid."""
        
        z = zaplet()

        block_re_str = '' # the block regex that will be built for this zaplet
        allow_re_str = '' # the allow regex that will be built for this zaplet

        z.lines = zapletlines

        in_zaplet = 0
        for line in z.lines:
            if in_zaplet:
                line = string.strip(line)
                if (len(line) == 0):
                    continue
                if re.search("\s*#", line):
                    pos = string.find(line, '#')
                    z.comment_list.append(line[pos+1:])
                    continue
                tokens = re.split('\s+', line)
                command = string.lower(tokens[0])
                if len(tokens) > 1:
                    arg = tokens[1]
                else:
                    arg = ""
                    
                if (command == 'version'):
                    vstr = arg
                    z.version = vstr
                    vlist = string.split(vstr,'.')
                    z.majorversion = string.atoi(vlist[0])
                    z.minorversion = string.atoi(vlist[1])
                    debug.debug(1,'zaplet version: %i.%i' % (z.majorversion, z.minorversion))
                elif (command == 'host'):
                    z.host = string.lower(arg)
                    debug.debug(1,'zaplet host: %s' % z.host)
                elif (command == 'block'):
                    z.block_list.append(arg)
                elif (command == 'allow'):
                    z.allow_list.append(arg)
                elif (command == '</zaplet>'):
                    in_zaplet = 0
                    continue
                else:
                    debug.debug(1,"unknown command: " + command)
                    pass
            else: # not in_zaplet
                if re.search('^<zaplet>',line):
                    in_zaplet = 1
                    continue

        # if there is no host, version, allow, or block statements,
        # then this is a default zaplet-- so fill in the fields automatically
        if (z.host == None and
            z.version == '' and
            z.allow_list == [] and
            z.block_list == []):

            z.host = zfile[:-4]  # get rid of the .zap
            z.version = "0.9"
            z.majorversion = 0
            z.minorversion = 9


        # if there is no version statement, don't use the zaplet--
        #   version statements are now required!
        if z.majorversion != None and z.minorversion != None:
            debug.debug(1,"z.version: %d.%d " % (z.majorversion,z.minorversion))
        else:
            debug.debug(1,"zaplet not used; no version statement!")
            return None

        # if there is no host statement, don't use the zaplet--
        #   host statements are now required
        if z.host != None:
            debug.debug(1,"z.host: %s " % z.host)
        else:
            debug.debug(1,"zaplet not used; no host statement!")
            return None


        # chop off initial |
        allow_re_str = allow_re_str[1:]
        block_re_str = block_re_str[1:]

        debug.debug(1,"allow_re_str: %s" % allow_re_str)
        debug.debug(1,"block_re_str: %s" % block_re_str)

        # compile regexes
        if allow_re_str == "":
            allow_re_str = "^$"
        z.allow = allow_re_str
        z.allow_re = re.compile(z.allow)

        z.block = block_re_str
        z.block_re = re.compile(z.block)

        return z



    def print_all_zaplets(self):
        """
        output all the zaplets in text format
        """
        zkeys=self.zaplet_list.keys()
        for zkey in zkeys:
            self.print_zaplet(self.zaplet_list[zkey])
        

    def print_zaplet(self,zaplet):
        """
        output one zaplet in text format
        """
        print
        print "<zaplet>"
        print "  version " + zaplet.version
        print "  host " + zaplet.host

        for allow in zaplet.allow_list:
            print "  allow " + allow
            
        for block in zaplet.block_list:
            print "  block " + block
        print "</zaplet>"




    def print_xml_all_zaplets(self):
        """
        output all the zaplets in text format
        """
        zkeys=self.zaplet_list.keys()
        for zkey in zkeys:
            self.print_xml_zaplet(self.zaplet_list[zkey])
        

    def xml_zaplet_repr(self,zaplet):
        """
        return string containing one zaplet in XML format
        """
        s = ""
        s = s + '<?xml version="1.0"?>\n'
        s = s + "<zaplet>\n"

        s = s + "  <version>0.9</version>\n"
        s = s + "  <host>" + zaplet.host + "</host>\n"

        if len(zaplet.comment_list) > 0:
            for comment in zaplet.comment_list:
                s = s + "  " + "<comment>" + comment + "</comment>\n"

        tmpstr = ""
        allows = 0
        for allow in zaplet.allow_list:
            tmpstr = "  <allow_url "
            expr = allow
            if regex_string(expr):
                tmpstr = tmpstr + 'type ="regex">' + expr[1:-1] + "</allow_url>\n"
            elif quoted_string(expr):
                tmpstr = tmpstr + 'type="literal">' + expr[1:-1] + "</allow_url>\n"
            else:
                tmpstr = tmpstr + 'type="literal">' + expr[1:-1] + "</allow_url>\n"
            if len(tmpstr) > 0:
                allows = allows + 1
                s = s + tmpstr
            
        tmpstr = ""
        blocks = 0
        for block in zaplet.block_list:
            tmpstr = "  <block "
            expr = block
            if regex_string(expr):
                tmpstr = tmpstr + 'type="regex">' + expr[1:-1] + "</block_url>\n"
            elif quoted_string(expr):
                tmpstr = tmpstr + 'type="literal">' + expr[1:-1] + "</block_url>\n"
            else:
                tmpstr = tmpstr + 'type="literal">' + expr[1:-1] + "</block_url>\n"
            if len(tmpstr) > 0:
                blocks = blocks + 1
                s = s + tmpstr

        if blocks == 0:
            s = s + '  <block_url type="everything"/>\n'
            
        s = s + "</zaplet>\n"

        return s

    def print_xml_zaplet(self,zaplet):
        """print out zaplet xml representation"""

        print self.xml_zaplet_repr(zaplet)


    def write_xml_zaplet(self,zaplet,output_dir):
        """
        output one zaplet in XML format
        """

        if len(output_dir) < 1:
            debug.debug(1,"empty output_dir")
        if output_dir[-1] != '/':
            output_dir = output_dir + '/'

        name = zaplet.host + ".zap"
        fullname = output_dir + name
        
        debug.debug(1,"writing file: %s" % fullname)
        
        try:
            file = open(fullname,'w')
        except:
            debug.debug(1,"could not open %s for writing" % fullname)
            return

        s = self.xml_zaplet_repr(zaplet)

        try:
            file.write(s)
        except:
            debug.debug(1,"could not write to %s" % fullname)


        try:
            file.close()
        except:
            debug.debug(1,"could not close %s" % fullname)

            
    def write_xml_all_zaplets(self,output_dir):
        """
        output all the zaplets in text format
        """

        # make output directory if it doesn't exist
        if os.path.exists(output_dir) != 1:
            debug.debug(1,"making output directory: %s" % output_dir)
            os.mkdir(output_dir)

        zkeys=self.zaplet_list.keys()
        for zkey in zkeys:
            self.write_xml_zaplet(self.zaplet_list[zkey],output_dir)
        
